from typing import Optional

import pkg_resources
from ultralytics import YOLO
from PIL import Image as Img

import torch


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
models = {}
brand_name = 'NIKE'
confidence = 0.5  # todo make config
output_images_dir = 'output'


def init_nike_model():
    # Load a model
    data_path = pkg_resources.resource_filename('nike_detector', 'models/')
    models['yolo_model'] = YOLO('yolov8x.yaml')  # build a new model from scratch
    models['yolo_model'] = YOLO('yolov8x.pt')  # load a pretrained model (recommended for training)
    models['yolo_model'] = YOLO(data_path + 'best_old.pt')
    # logger.info("Trained YOLO model is successfully loaded on: {device}".format(device=device))

    return models['yolo_model']


def predict(image,
            save=True,
            save_txt=True):  # path to img or url
    run_name = 'predict'
    results = models['yolo_model'].predict(image,
                                           conf=confidence,
                                           device=device,
                                           save=True if save else False,
                                           project=output_images_dir,
                                           name=run_name,
                                           save_txt=True if save_txt else False,
                                           exist_ok=True)

    result = results[0]
    print(result)
    im_array = result.plot()  # plot a BGR numpy array of predictions
    im_array = im_array[..., ::-1]
    im = Img.fromarray(im_array)  # RGB PIL image

    results = {
        "image": im,
        "image_size": {
            "height": im_array.shape[0],
            "width": im_array.shape[1],
            "channels": im_array.shape[2],
        },
        "boxes": result.boxes.xywhn.tolist()[0]
    }

    return results


def track(video_path: Optional[str],
          save=True,
          save_txt=True):  # path to img or url
    run_name = 'track'

    result = models['yolo_model'].track(source=video_path,
                                        conf=confidence,
                                        device=device,
                                        save=True,
                                        project=output_images_dir,
                                        name=run_name,
                                        save_txt=True,
                                        exist_ok=True)

    results = {
        'video_save_path': output_images_dir + '/' + run_name + '/' + video_path,
        'result': result
    }

    return results



import asyncio
import aiohttp
import io
from hashlib import md5
import logging
import numpy as np
from datetime import datetime, timedelta

from PIL import Image
from typing import List, Optional, Dict, Union, Tuple, Any

from ha_core.utils.measurer import track
from ha_core.utils.misc import chunks
from ha_core.config.base import CDN_IP, CDN_SALT, CDN_URL, IS_PRODUCTION
from ha_core.utils import safe_time


N_RETRIES = 1
COOLDOWN = 0
BATCH_SIZE = 500
SUCCESS_RESPONSE_CODE = 200


class DaemonBaseAsync:
    def __init__(self):
        self.session = aiohttp.ClientSession()

    def __del__(self):
        asyncio.get_event_loop().run_until_complete(self.session.close())


class DaemonRestDataAsync(DaemonBaseAsync):
    @track('daemon_rest_async')
    async def get_by_url(
        self, url: str, tag: str, source=None, retries: int = 1, cooldown: Optional[int] = None
    ) -> Tuple[str, int, Any]:
        for _ in range(retries):
            resp = await self.session.request(
                method='GET',
                url=url,
                headers={'User-Agent': 'ds:' + (source or 'rest_service')},
                verify_ssl=False,
            )
            result, status = await resp.read(), resp.status
            if status == 425 and cooldown:
                await asyncio.sleep(cooldown)

        return tag, status, result


def _get_sign(url_path, params, timestamp):
    for key in sorted(params):
        url_path += key + params[key]
    return md5((timestamp + url_path + CDN_SALT).encode()).hexdigest()


def get_cdn_link(
        file_path=None,
        social_name=None,
        social_item=None,
        social_item_id=None,
        width=None,
        height=None,
        till=None,
        use_ip=False,
        longpoll=False,
        dernd=False
):
    """Returns link for HA CDN: file from storage or user avatar or post image

    :param str|None file_path: path for the file from HA storage
    :param str|None social_name: instagram, tiktok, etc
    :param str|None social_item: user or post
    :param str|None social_item_id: user/account social id or post's code
    :param int|None width: width of the image if resize is needed, use only width or height for keeping aspect ratio
    :param int|None height: height of the image if resize is needed, use only width or height for keeping aspect ratio
    :param datetime|None till: valid until datetime
    :param bool use_ip: generate link with ip instead of domain for internal usage

    :return: cdn link
    :rtype: str
    """

    # костыль чтобы ссылки не тухли за час
    till = datetime.utcnow() + timedelta(hours=1)
    timestamp = safe_time.to_timestamp(till or datetime.utcnow())
    timestamp = str(timestamp - timestamp % 3600)

    params = {}
    if width:
        params['w'] = str(width)
    if height:
        params['h'] = str(height)
    if till:
        params['till'] = timestamp

    if file_path:
        url_path = '/ha/' + file_path
    else:
        url_path = f'/img/{social_name}/{social_item}/{social_item_id}.jpg'

    params['sign'] = _get_sign(url_path, params, timestamp)

    if longpoll:
        params['longpoll'] = '1'
    if dernd:
        params['dernd'] = '1'

    host = CDN_URL if not use_ip or not IS_PRODUCTION else CDN_IP

    return host + url_path + '?' + '&'.join(['='.join(i) for i in params.items()])


def get_ig_avatar(social_id: str, big=False, use_ip=False, longpoll=False, dernd=False) -> str:
    if not social_id:
        return ''

    return get_cdn_link(
        social_name='instagram',
        social_item='user',
        social_item_id=social_id,
        width=320 if big else 150,
        use_ip=use_ip,
        longpoll=longpoll,
        dernd=dernd
    )


def m_get_image_contents_with_code(
    urls: Dict[Union[str, int], str],
    source: Optional[str],
    batch_size: int,
    retries: int,
    cooldown: int,
) -> dict:
    daemon_rest_async = DaemonRestDataAsync()
    tasks = [
        daemon_rest_async.get_by_url(url, u_id, source or 'cdn_service', retries, cooldown)
        for u_id, url in urls.items()
    ]
    res = {}
    for chunk in chunks(tasks, batch_size):
        for complete_task in asyncio.get_event_loop().run_until_complete(
            asyncio.gather(*chunk, return_exceptions=True)
        ):
            res[complete_task[0]] = {'status': complete_task[1], 'bin': complete_task[2]}
    return res


@track('get_avatars')
def get_avatars(user_ids: List[str]):
    images = [None for _ in user_ids]
    codes = [-1 for _ in user_ids]

    img_urls = {
        # user_id: get_ig_avatar(user_id, big=True, dernd=True if not precise_detection else False)
        user_id: get_ig_avatar(user_id, big=True, dernd=True)
        for user_id in user_ids
    }
    imgs = m_get_image_contents_with_code(img_urls, 'new_agre_detector', BATCH_SIZE, N_RETRIES, COOLDOWN)

    for i, (user_id, data) in enumerate(imgs.items()):
        if data['status'] == 200:
            if data['bin']:
                img_ = Image.open(io.BytesIO(data['bin']))
                img = np.asarray(img_)
                images[i] = img

            else:
                logging.debug(f"Can't load avatar for account: {user_id}")
                images[i] = None

        elif data['status'] == 425:
            logging.debug("425 code from CDN")

        elif data['status'] == 410:
            logging.debug("410 code from CDN")

        elif data['status'] == 417:
            logging.debug("417 code from CDN")

        codes[i] = data['status']

    return list(img_urls.values()), images, codes


@track('get_ig_post')
def get_ig_post(post_id: str, use_ip=False, longpoll=False, dernd=False) -> str:
    if not post_id:
        return ''

    return get_cdn_link(
        social_name='instagram',
        social_item='post',
        social_item_id=post_id,
        width=480,
        use_ip=use_ip,
        longpoll=longpoll,
        dernd=dernd,
    )


@track('get_post_images')
def get_post_images(posts: List[str]):
    images = {}
    codes = {}
    img_urls = {post: get_ig_post(post, dernd=True) for post in posts}
    imgs = m_get_image_contents_with_code(img_urls, 'new_agre_detector', BATCH_SIZE, N_RETRIES, COOLDOWN)

    for post_id, data in imgs.items():
        if data['status'] == 200:
            if data['bin']:
                img_ = Image.open(io.BytesIO(data['bin']))
                # img = np.asarray(img_)
                images[post_id] = img_
                codes[post_id] = data['status']
            else:
                images[post_id] = None
                codes[post_id] = data['status']

        elif data['status'] == 425:
            images[post_id] = None
            codes[post_id] = data['status']

        elif data['status'] == 410:
            images[post_id] = None
            codes[post_id] = data['status']
        else:
            images[post_id] = None
            codes[post_id] = data['status']

    return img_urls, images, codes



import os
from typing import Dict

from ha_core.utils.db import get_hz_db

from constants import LIST_OF_BRANDS, LIST_OF_BRANDS_REVERSED, LIST_OF_BRAND_ACCOUNTS
from dataset.avatars import get_post_images


# def get_user_ids_for_brand(brand_name):
#     user_ids = []
#     cursor = get_db()['instagram_report_data'].find({}, {'_id': True, 'basic.description': True}).limit(1000000)
#     for item in cursor:
#         if brand_name in item['basic']['description'] or brand_name.title() in item['basic']['description']:
#             user_ids.append(item['_id'])
#
#     return user_ids


def get_post_ids_for_brand(brand_name):
    post_ids = []
    cursor = get_hz_db('instagram')['instagram_post'].find({}, {'_id': True, 'caption': True}).limit(1000000)  # todo check collection
    for item in cursor:
        if brand_name in item.get('caption', '').split() or \
                brand_name.title() in item.get('caption', '').split() or \
                '#' + brand_name in item.get('caption', '').split() or \
                '#' + brand_name.lower() in item.get('caption', '').split():
            post_ids.append(item['_id'])

    return post_ids


def get_premium_post_ids_for_brand(brand_name):
    post_ids = []
    brand_account_ids = LIST_OF_BRAND_ACCOUNTS[LIST_OF_BRANDS_REVERSED[brand_name]]
    cursor = get_hz_db('instagram')['instagram_post'].find({'user_id': {'$in': [str(_id) for _id in brand_account_ids]}}, {'_id': True, 'caption': True}).limit(100000)  # todo check collection
    for item in cursor:
        post_ids.append(item['_id'])

    return post_ids


def save_images_to_directory(directory_name,
                             images_by_post_ids):
    valid_post_ids = []

    # Parent Directory path
    parent_dir = '/Users/mikhail_korotkov/PycharmProjects/brand_detection/BrandDetection/dataset'

    # Path
    path = os.path.join(parent_dir, 'images')
    if not os.path.exists(path):
        os.mkdir(path)
    path = os.path.join(path, directory_name)
    os.mkdir(path)
    print(path)
    path = os.path.join(parent_dir, 'labels')
    if not os.path.exists(path):
        os.mkdir(path)
    path = os.path.join(path, directory_name)
    os.mkdir(path)
    # os.mkdir('labels/' + directory_name + '/')

    for post_id, image in images_by_post_ids.items():
        if image:
            valid_post_ids.append(post_id)
            save_image(directory_name, post_id, image)

    if valid_post_ids:
        with open('labels/' + directory_name + '/' + 'labels.txt', 'w') as f:
            for post_id in valid_post_ids:
                # write each item on a new line
                f.write("%s\n" % post_id + '.jpg')


def save_image(directory_name, post_id, image):
    image.save('images/' + directory_name + '/' + str(post_id) + '.jpg')


def get_post_ids_for_brands(brand_names: Dict[int, str]):
    for _id, brand_name in brand_names.items():
        print(brand_name)
        post_ids1 = get_post_ids_for_brand(brand_name=brand_name)
        post_ids2 = get_premium_post_ids_for_brand(brand_name=brand_name)
        post_ids = post_ids1 + post_ids2
        print(len(post_ids), post_ids[0])
        img_urls, images_by_post_ids, codes = get_post_images(posts=list(set((post_ids))))
        save_images_to_directory(directory_name=brand_name,
                                 images_by_post_ids=images_by_post_ids)


get_post_ids_for_brands(LIST_OF_BRANDS)
